﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Net.Sockets;
using System.IO;

namespace TestInetRadio {
    public partial class Form1: Form {
        // 音源リクエストのタイプ
        // 直接音源を要求する(一度GETしたら連続的に音源を取得可能)。プレイリストを介して要求する(小刻みに音源を要求する)。
        private enum REQUEST_TYPE { DIRECT, PLAYLIST };    

        // HTTPレスポンスのボディ部の音源データのフォーマット
        private enum TRANSPORT_FORMAT { MP3, AAC, TS_AAC };

        private class StationInfo {
            public String station;      // 局名
            public REQUEST_TYPE request_type;
            public TRANSPORT_FORMAT transport_format;       // MP3, TS
            public String url;
        };

        private class UrlInfo {
            //public String station;      // 局名
            //public REQUEST_TYPE request_type;
            //public TRANSPORT_FORMAT transport_format;       // MP3, TS
            //public String url;
            //public String location;

            public String host;
            public String port;
            public String file;
            //public String playlist;
        };

        private class HttpHeaderInfo {
            public bool Http;               // HTTPレスポンスヘッダが含まれているときtrue。
            public bool HeaderComplete;     // HTTPレスポンスヘッダがすべて取得（受信）できているときtrue。
            public String status;
            public String ContentType;
            public String Location;
            public int ContentLength;
            public int DataIndex;           // HTTPレスポンス中のエンティティ開始位置(0のときヘッダなしで、エンティティの続き)

            public HttpHeaderInfo()
            {
                HeaderComplete = false;
                status = "";
                ContentType = "";
                Location = "";
                ContentLength = 0;
                DataIndex = 0;  // -1;
                //Length = 0;
            }
        };

        private class M3Utable {
            public int duration;        // #EXTINF:の値を10ms(timer1.Interval)単位で
            public String reqfile;      // ファイル名またはURL
        }

        private TcpClient tcp;
        private NetworkStream stream;
        private FileStream outStream;
        private FileStream sndStream;
        private FileStream sndtsStream;
        private FileStream sndStream2;
        private int SumDuration;
        private bool ReceiveStart = false;
        private DateTime StartDt;
        private DateTime LastGetDt;

        private StationInfo CurrentStation;
        private UrlInfo CurrentUrl;
        private byte[] RecvBuf = new byte[RECV_BUF_SIZE];
        private int RecvBufBytes;
        private int ContentLength = 0;
        private int ContentGetByte = 0;
        private HttpHeaderInfo HeaderInfo = new HttpHeaderInfo();

        private const String ROOT_DIR = @"C:\UserData\Source\Ele\TEST_テスト\TestInetRadio\";
        // HTTPレスポンスヘッダも含めて全ての取得データを保存
        private const String OUTFILE = ROOT_DIR + "alldata.dat";
        // ボディ部から抽出したMP3データの出力先
        private const String DATFILE_MP3 = ROOT_DIR + "sound.mp3";
        // ボディ部から抽出したMP3データの出力先
        private const String DATFILE_AAC = ROOT_DIR + "sound.aac";
        // ボディ部から抽出したTSデータの出力先
        private const String DATFILE_TS = ROOT_DIR + "sound.ts";
        // ボディ部から抽出した(その他)データの出力先
        private const String DATFILE = ROOT_DIR + "sound.dat";
        // TSフォーマットからの抽出音源（リアルタイム処理）
        private const String SNDFILE_TS_AAC = ROOT_DIR + "soundts.aac";
        // TSファイルからの抽出音源（事後処理）
        private const String SNDFILE2 = ROOT_DIR + "sound2.aac";
        //
        private int FragmentNo = 0;

        private const int RECV_BUF_SIZE = 1024 * 2;           // 最低、PLAYLISTを１回で取得できるサイズ
        private const int TS_SIZE = 188;

        private const int M3U_TABLE_SIZE = 30;
        private M3Utable[] M3uTable = new M3Utable[M3U_TABLE_SIZE];
        private int NumM3uTable = 0;            // M3uTable内の有効項目数
        private int M3uTableInsIndex = 0;       // M3uTable内の新規項目挿入位置
        private int M3uTableGetIndex = 0;       // M3uTable内の次回音源データ取得対象位置
        private int NextGetTimer = -1;          // 音源データリクエスト開始までのタイマーカウンタ
        private int CurrentDuration = 0;        // 現時点の音源データリクエストのduration
        private bool M3uRequestFlag = false;
        private int TickCountAtRead = 0;

        private byte[] TsWork = new byte[TS_SIZE];  // TSワーク変数
        private int TsWorkBytes = 0;
        int TsPmtPid = 0;
        int TsElementaryPidSound = 0;
        private const int PID_NULL = 0x1FFF;        // 8191, NULLパケット

        // GETで要求しているデータのタイプ
        private enum GET_DATA_TYPE { NONE, SOUND, M3U };                // なし、音源、M3Uプレイリスト
        private GET_DATA_TYPE TargetDataTypeOfGet = GET_DATA_TYPE.NONE;

        // program clock reference取得
        private int TickCountAtPcr = 0;
        private ulong Pcr = 0;
        private byte[] PcrByte = new byte[6];
        private bool PcrStart = false;
        private int SumPcrMsec = 0;
        private int SumTcMsec = 0;

        public Form1()
        {
            InitializeComponent();
        }

        /// <summary>
        /// 受信開始
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnConnect_Click(object sender, EventArgs e)
        {
            try {
                if (cboUrl.SelectedIndex < 0) return;
                CurrentStation = GetStationInfo(cboUrl.Text);
                //CurrentUrl = GetUrlInfo(cboUrl.Text);
                //if (CurrentUrl == null) return;

                // outStream : HTTPレスポンスすべて（ヘッダ＋メッセージボディ＋デバック情報）
                outStream = new FileStream(OUTFILE, FileMode.Create);
                // sndStream : HTTPレスポンスのメッセージボディの内、プレイリストを除くバイナリデータすべて。
                if (CurrentStation.transport_format == TRANSPORT_FORMAT.MP3) {
                    sndStream = new FileStream(DATFILE_MP3, FileMode.Create);
                } else if (CurrentStation.transport_format == TRANSPORT_FORMAT.AAC) {
                    sndStream = new FileStream(DATFILE_AAC, FileMode.Create);
                } else if (CurrentStation.transport_format == TRANSPORT_FORMAT.TS_AAC) {
                    sndStream = new FileStream(DATFILE_TS, FileMode.Create);
                    sndtsStream = new FileStream(SNDFILE_TS_AAC, FileMode.Create);      // TSデータからリアルタイムで音声データを抽出
                } else {
                    sndStream = new FileStream(DATFILE, FileMode.Create);
                }
                SumDuration = 0;
                NextGetTimer = -1;
                TsWorkBytes = 0;
                M3uTableInsIndex = 0;

                // URLに対して最初のGETリクエスト
                ConnectAndFirstGet(CurrentStation.url);
                ReceiveStart = true;
                PcrStart = false;

                StartDt = DateTime.Now;
                lblStartTime.Text = StartDt.ToString("HH:mm:ss");

                return;

            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
            } finally {
            }
        }

        /// <summary>
        /// GETリクエスト送信
        /// </summary>
        private void SendGetRequest(String host, String targetfile)
        {
            String msg = "GET " + targetfile + " HTTP/1.1\r\n"
                + "HOST: " + host + "\r\n"
                + "Accept: */*\r\n"
                + "\r\n";

            System.Diagnostics.Debug.WriteLine("host=" + host + " file=" + targetfile);

            try {
                Byte[] sendBytes = Encoding.ASCII.GetBytes(msg);    // Encoding.UTF8.GetBytes(msg);
                stream.Write(sendBytes, 0, sendBytes.Length);

                //if (targetfile.Length > 5 && targetfile.Substring(targetfile.Length - 5, 5).Equals(".m3u8", StringComparison.OrdinalIgnoreCase) == true) {
                if (targetfile.Contains(".m3u8") == true) {
                    TargetDataTypeOfGet = GET_DATA_TYPE.M3U;
                } else {
                    TargetDataTypeOfGet = GET_DATA_TYPE.SOUND;
                }
            } catch (Exception ex) {
                System.Diagnostics.Debug.WriteLine("SendGetRequest:" + ex.Message);
            }

            byte[] b = Encoding.UTF8.GetBytes("\r\nSendGetRequest=" + targetfile + "\r\n");
            outStream.Write(b, 0, b.Length);
        }

        /// <summary>
        /// StationInfo取得
        /// </summary>
        /// <param name="text">DropDownListの選択値</param>
        /// <returns></returns>
        private StationInfo GetStationInfo(String text)
        {
            StationInfo info = new StationInfo();
            int idx1, idx2;

            idx1 = text.IndexOf(':');
            if (idx1 < 0) return null;
            idx2 = text.IndexOf(':', idx1 + 1);
            if (idx2 < 0) return null;

            info.station = text.Substring(0, idx1);
            String stype = text.Substring(idx1 + 1, idx2 - idx1 - 1);
            if (stype == "TS_AAC") {
                info.transport_format = TRANSPORT_FORMAT.TS_AAC;
            } else if (stype == "MP3") {
                info.transport_format = TRANSPORT_FORMAT.MP3;
            } else if (stype == "AAC") {
                info.transport_format = TRANSPORT_FORMAT.AAC;
            } else {
                info.transport_format = TRANSPORT_FORMAT.MP3;
            }
            info.url = text.Substring(idx2 + 1);

            if (info.url.Contains(".m3u8")){
                info.request_type = REQUEST_TYPE.PLAYLIST;
            } else {
                info.request_type = REQUEST_TYPE.DIRECT;
            }

            return info;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="text">DropDownListの選択値</param>
        /// <returns>UrlInfo</returns>
        //private UrlInfo GetUrlInfo(String text)
        //{
        //    UrlInfo info = new UrlInfo();
        //    int idx1, idx2;

        //    idx1 = text.IndexOf(':');
        //    if (idx1 < 0) return null;
        //    idx2 = text.IndexOf(':', idx1 + 1);
        //    if (idx2 < 0) return null;

        //    info.station = text.Substring(0, idx1);
        //    String stype = text.Substring(idx1 + 1, idx2 - idx1 - 1);
        //    if (stype == "TS_AAC") {
        //        info.transport_format = TRANSPORT_FORMAT.TS_AAC;
        //    } else if (stype == "MP3") {
        //        info.transport_format = TRANSPORT_FORMAT.MP3;
        //    } else if (stype == "AAC") {
        //        info.transport_format = TRANSPORT_FORMAT.AAC;
        //    } else {
        //        info.transport_format = TRANSPORT_FORMAT.MP3;
        //    }
        //    String url = text.Substring(idx2 + 1);

        //    String[] urlpart = DevideUrl(url);
        //    info.url = url;
        //    info.host = urlpart[0];
        //    info.port = urlpart[1];
        //    String file = urlpart[2];

        //    if (file.Length > 5 && file.Substring(file.Length - 5, 5).Equals(".m3u8", StringComparison.OrdinalIgnoreCase) == true) {
        //    //if (info.file.Length > 5 && info.file.Substring(info.file.Length - 5, 5).Equals(".m3u8", StringComparison.OrdinalIgnoreCase) == true) {
        //        info.request_type = REQUEST_TYPE.PLAYLIST;
        //        info.playlist = urlpart[2];
        //    } else {
        //        info.request_type = REQUEST_TYPE.DIRECT;
        //        info.file = urlpart[2];
        //    }

        //    return info;
        //}

        private HttpHeaderInfo GetHttpHeaderInfo(byte[] bytes, int length)
        {
            int lineno = 1;
            int idx1, idx2;
            String field;
            String val;
            HttpHeaderInfo info = HeaderInfo;   // new HttpHeaderInfo();

            if (!(bytes[0] == 'H' && bytes[1] == 'T' && bytes[2] == 'T' && bytes[3] == 'P')) {
                info.Http = false;
                info.DataIndex = 0;
                return info;
            }
            info.Http = true;
            info.HeaderComplete = false;
            info.status = "";
            info.ContentType = "";
            info.Location = "";
            info.ContentLength = 0;

            String response = Encoding.UTF8.GetString(bytes, 0, length);

            try {
                idx1 = 0;
                while (true) {
                    if (lineno == 1) {
                        if ((idx2 = response.IndexOf(' ', idx1)) == -1) break;
                        info.status = response.Substring(idx2 + 1, 3);
                        if ((idx2 = response.IndexOf('\n', idx1)) == -1) break;
                        idx1 = idx2 + 1;
                    } else {
                        if (bytes[idx1] == '\r') {          // 空行(HTTPヘッダの終わり)
                            info.DataIndex = idx1 + 2;
                            info.HeaderComplete = true;
                            break;
                        }
                        if ((idx2 = response.IndexOf(':', idx1)) == -1) break;
                        field = response.Substring(idx1, idx2 + 1 - idx1);
                        idx2++;
                        while (response[idx2] == ' ') idx2++;
                        if ((idx1 = response.IndexOf('\r', idx1)) == -1) break;
                        val = response.Substring(idx2, idx1 - idx2);
                        idx1 += 2;

                        if (field == "Location:") {
                            info.Location = val;
                        } else if (field == "Content-Type:") {
                            info.ContentType = val;
                        } else if (field == "Content-Length:") {
                            try {
                                info.ContentLength = Int32.Parse(val);
                            } catch (Exception ex) {
                                info.ContentLength = 0;
                            }
                        }
                        //System.Diagnostics.Debug.WriteLine("info field=" + field + " val=" + val);
                    }

                    lineno++;
                }
            } catch (Exception ex) {

            }

            return info;
        }


        /// <summary>
        /// M3U playlistの解析
        /// </summary>
        /// <param name="bytes"></param>
        /// <param name="start_index"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        private bool ParseM3U(byte[] bytes, int start_index, int length)
        {
            StreamReader reader = new StreamReader(new MemoryStream(bytes, start_index, length));

            String line = null;
            int idx1, idx2, idx3;
            int i;
            int duration = 0;

            while ((line = reader.ReadLine()) != null) {
                if (line.StartsWith("#EXTINF:", StringComparison.OrdinalIgnoreCase) == true) {
                    char[] buf = new char[10];
                    int idx = 0;

                    idx1 = line.IndexOf(':');
                    idx1++;
                    while (!(line[idx1] >= '0' && line[idx1] <= '9')) {
                        idx1++;
                    }
                    idx2 = idx1;        // idx1は数値部分始まり
                    while (line[idx2] >= '0' && line[idx2] <= '9') {
                        buf[idx] = line[idx2];
                        idx++;
                        idx2++;
                    }
                    if (line[idx2] == '.') {
                        idx3 = idx2 + 1;
                        for (i=0; i<3; i++) {   // 小数点以下を３ケタ取得
                            if (line[idx3] >= '0' && line[idx3] <= '9') {
                                buf[idx++] = line[idx3++];
                            } else {
                                buf[idx++] = '0';
                            }
                        }
                        //M3uTable[NumM3uTable].duration = Int32.Parse(new String(buf, 0, idx));
                        duration = Int32.Parse(new String(buf, 0, idx));
                        //M3uTable[LastM3uTableIndex].duration = Int32.Parse(new String(buf, 0, idx));
                    } else {
                        //M3uTable[NumM3uTable].duration = Int32.Parse(new String(buf, 0, idx)) * 100;
                        duration = Int32.Parse(new String(buf, 0, idx)) * 1000;
                        //M3uTable[LastM3uTableIndex].duration = Int32.Parse(new String(buf, 0, idx)) * 100;
                    }

                    //System.Diagnostics.Debug.WriteLine("#EXTINF=" + String.Format("{0}", M3uTable[NumM3uTable].duration));
                    //System.Diagnostics.Debug.WriteLine("#EXTINF=" + String.Format("{0}", M3uTable[LastM3uTableIndex].duration));
                    //System.Diagnostics.Debug.WriteLine("#EXTINF=" + String.Format("{0}", duration));
                } else if (line.Length > 0 && line[0] != '#') {
                    for (i=0; i< M3U_TABLE_SIZE; i++) {
                        if (M3uTable[i].reqfile == line) break;
                    }
                    if (i == M3U_TABLE_SIZE /*&& duration > 0*/) {
                        System.Diagnostics.Debug.WriteLine(String.Format("{0}", M3uTableInsIndex) + " #EXTINF=" + String.Format("{0}", duration) + "," + line);
                        M3uTable[M3uTableInsIndex].duration = duration;
                        M3uTable[M3uTableInsIndex].reqfile = line;
                        duration = 0;
                        M3uTableInsIndex++;
                        if (M3uTableInsIndex == M3U_TABLE_SIZE) {
                            M3uTableInsIndex = 0;
                        }
                        NumM3uTable++;
                        if (NumM3uTable == M3U_TABLE_SIZE) {
                            break;
                        }
                    }
                }
            }
            System.Diagnostics.Debug.WriteLine("NumM3uTable=" + NumM3uTable.ToString());

            return true;
        }


        /// <summary>
        /// 10ms interval
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void timer1_Tick(object sender, EventArgs e)
        {
            int read_byte;

            if (ReceiveStart == true) {         // 最初のGETを送信済み＝受信開始
                while (stream.DataAvailable == true && RecvBufBytes < RECV_BUF_SIZE) {
                    // 受信はすべてここで行う
                    read_byte = stream.Read(RecvBuf, RecvBufBytes, RECV_BUF_SIZE - RecvBufBytes);
                    outStream.Write(RecvBuf, RecvBufBytes, read_byte);
                    RecvBufBytes += read_byte;
                    TickCountAtRead = Environment.TickCount;
                }
                if (RecvBufBytes > 0) {
                    //System.Diagnostics.Debug.WriteLine(" RecvBufBytes=" + RecvBufBytes.ToString());
                    if (RecvBufBytes == RECV_BUF_SIZE || (Environment.TickCount - TickCountAtRead) > 10) {
                        HttpHeaderInfo resheader = GetHttpHeaderInfo(RecvBuf, RecvBufBytes);
                        if (resheader.Http == true && resheader.HeaderComplete == true) {
                            System.Diagnostics.Debug.WriteLine("HttpHeaderInfo " + resheader.Location
                                + " status=" + resheader.status
                                + " ContentType=" + resheader.ContentType
                                + " ContentLength=" + resheader.ContentLength
                                + " dataindex=" + resheader.DataIndex);
                            ContentLength = resheader.ContentLength;
                            ContentGetByte = 0;

                            if (/*resheader.Http == true &&*/ resheader.status.Length > 0) {
                                if (resheader.status[0] == '3') {
                                    System.Diagnostics.Debug.WriteLine("status=" + resheader.status + " Location=" + resheader.Location);
                                    ReconnectAndFirstGet(resheader.Location);       // LocationがPLAYLISTであることは想定外
                                    RecvBufBytes = 0;
                                    return;
                                } else if (resheader.status[0] != '2') {
                                    System.Diagnostics.Debug.WriteLine("status code 不正 " + resheader.status);
                                    ReceiveStart = false;
                                    return;
                                }
                            }
                        } else if (resheader.Http == true && resheader.HeaderComplete == false) {
                            // HTTPヘッダをすべて受信できていない
                            return;
                        }

                        if (TargetDataTypeOfGet == GET_DATA_TYPE.SOUND) {
                            //System.Diagnostics.Debug.WriteLine(" RecvBufBytes=" + RecvBufBytes.ToString() + " SOUND");
                            int content_size = RecvBufBytes - resheader.DataIndex;
                            ContentGetByte += content_size;

                            if (CurrentStation.transport_format == TRANSPORT_FORMAT.MP3 || CurrentStation.transport_format == TRANSPORT_FORMAT.AAC) {
                                sndStream.Write(RecvBuf, resheader.DataIndex, content_size);
                                if (sndStream2 != null) sndStream2.Write(RecvBuf, resheader.DataIndex, content_size);
                                RecvBufBytes = 0;
                            } else if (CurrentStation.transport_format == TRANSPORT_FORMAT.TS_AAC) {
                                int i;
                                int zan1;       // TsWorkの空きバイト数
                                int zan2;       // RecvBufの未処理バイト数
                                int cpsize;     // RecvBufからTsWorkへのコピーバイト数
                                bool x47_err = false;

                                sndStream.Write(RecvBuf, resheader.DataIndex, content_size);
                                if (sndStream2 != null) sndStream2.Write(RecvBuf, resheader.DataIndex, content_size);
                                i = resheader.DataIndex;
                                while (i < RecvBufBytes) {
                                    if (TsWorkBytes == 0) {
                                        // 頭出し。0x47から始まるはずだが、ごくまれにずれている。
                                        if (RecvBuf[i] != 0x47) {       // 先頭バイト
                                            x47_err = true;
                                            i++;
                                            continue;
                                        }
                                    }
                                    if (x47_err == true) {
                                        System.Diagnostics.Debug.WriteLine("0x47 error");
                                    }
                                    // TsWorkに188バイト取り出す
                                    zan1 = (TS_SIZE - TsWorkBytes);
                                    zan2 = (RecvBufBytes - i);
                                    cpsize = (zan1 < zan2 ? zan1 : zan2);       // 両者の小さい方
                                    Array.Copy(RecvBuf, i, TsWork, TsWorkBytes, cpsize);
                                    TsWorkBytes += cpsize;
                                    i += cpsize;

                                    if (TsWorkBytes == TS_SIZE) {
                                        // TSパケット処理
                                        ParseTsWork(sndStream.Position);
                                        TsWorkBytes = 0;
                                        //Array.Clear(TsWork, 0, TS_SIZE);
                                    }
                                    x47_err = false;
                                }
                                RecvBufBytes = 0;
                            }

                            if (ContentGetByte == ContentLength) {
                                M3uRequestFlag = true;
                            }
                        } else if (TargetDataTypeOfGet == GET_DATA_TYPE.M3U) {
                            int content_size = RecvBufBytes - resheader.DataIndex;
                            ContentGetByte += content_size;
                            if ((RecvBufBytes - resheader.DataIndex) == ContentLength) {
                                System.Diagnostics.Debug.WriteLine(" M3U 一括取得 (Read bytes=" + RecvBufBytes.ToString() + ")");
                                // M3Uファイルの解析
                                if (ParseM3U(RecvBuf, resheader.DataIndex, RecvBufBytes - resheader.DataIndex) == false) {
                                    throw new Exception("M3Uファイル不正");
                                } else if (M3uTableInsIndex == 1 && M3uTable[0].reqfile.Contains(".m3u8") == true) {
                                    // M3Uファイルの記述が再びM3Uファイルの場合(コミュニティFM)。
                                    // （M3Uファイルが複数記載されていたら？）　i-waveの場合１件で、以後これをM3Uファイルとする。
                                    ReconnectAndFirstGet(M3uTable[0].reqfile);
                                } else {
                                    // M3Uファイルに記述された最初の音源を要求
                                    if (NextGetTimer < 0) {
                                        NextGetTimer = 0;
                                    }
                                }
                                RecvBufBytes = 0;
                            } else {    // すべてのM3Uを受信出来ていない（時々発生する）
                                // 再GETしない。時間をかければたいていすべて受信できる。
                                // すべて受信できなくても、時間が来たら(NextGetTimer)次の音源取得へ進む。そのあとでM3UのGET。
                            }
                        }
                    }
                }

                if (CurrentStation.request_type == REQUEST_TYPE.PLAYLIST) {
                    // PLAYLISTの音源を一つ取得するたびにPLAYLISTを要求する
                    if (M3uRequestFlag == true) {
                        SendGetRequest(CurrentUrl.host, CurrentUrl.file);
                        RecvBufBytes = 0;
                        M3uRequestFlag = false;
                    }

                    // 次の音源取得
                    if (NextGetTimer > 0) {
                        NextGetTimer -= timer1.Interval;
                        if (NextGetTimer < 0) NextGetTimer = 0;
                    }
                    if ((CurrentDuration > 0 && (DateTime.Now - LastGetDt).TotalMilliseconds >= CurrentDuration) || NextGetTimer == 0) { 
                    //if (NextGetTimer == 0) {
                        //NextGetTimer = -1;
                        if (NumM3uTable > 0) {
                            //if (ContentLength != ContentGetByte) {
                            //    // 直前の取得は完了していない。
                            //    System.Diagnostics.Debug.WriteLine(" TIMEOUT ContentLength=" + ContentLength.ToString() + " getbytes=" + ContentGetByte.ToString());
                            //}
                            if (ContentLength == ContentGetByte) {
                                NextGetTimer = -1;
                                SendM3uSoundRequest();
                                RecvBufBytes = 0;
                                M3uTableGetIndex++;
                                if (M3uTableGetIndex == M3U_TABLE_SIZE) {
                                    M3uTableGetIndex = 0;
                                }
                                NumM3uTable--;
                            }
                        } else {
                            NextGetTimer = -1;
                            System.Diagnostics.Debug.WriteLine("Empty M3uTable PlayM3uTableIndex=" + M3uTableGetIndex.ToString());
                            SendGetRequest(CurrentUrl.host, CurrentUrl.file);
                            RecvBufBytes = 0;
                            M3uRequestFlag = false;
                        }

                    }
                }
            }
        }


        /// <summary>
        /// M3Uファイルに従って音源データを要求する
        /// </summary>
        private void SendM3uSoundRequest()
        {
            String targetfile;
            int idx1;

            String[] urlpart = DevideUrl(M3uTable[M3uTableGetIndex].reqfile);
            if (urlpart[0] == "") {     // URLではない
                // PLAYLISTファイルと同じパスであることを前提
                idx1 = CurrentUrl.file.LastIndexOf('/');
                targetfile = CurrentUrl.file.Substring(0, idx1 + 1) + M3uTable[M3uTableGetIndex].reqfile;
                SendGetRequest(CurrentUrl.host, targetfile);
            } else {    // URL (host/portは同じである前提)
                if (CurrentUrl.host == urlpart[0] && CurrentUrl.port == urlpart[1]) {
                    SendGetRequest(CurrentUrl.host, urlpart[2]);
                } else {
                    throw new Exception("M3Uファイルに記述されたファイルのHOSTがM3Uファイル自体のHOSTと異なる");
                }
            }
            NextGetTimer = M3uTable[M3uTableGetIndex].duration - (1 * 100);    // -1sec TODO:最適値は？
            CurrentDuration = M3uTable[M3uTableGetIndex].duration;
            //NextGetTimer = M3uTable[M3uTableGetIndex].duration / 2;    // -1sec TODO:最適値は？
            //if (PlayM3uTableIndex == LastM3uTableIndex) {
            //    NextGetTimer = M3uTable[PlayM3uTableIndex].duration / 2;            //  TODO:最適値は？
            //} else {
            //    NextGetTimer = M3uTable[PlayM3uTableIndex].duration - (2 * 100);    // -2sec TODO:最適値は？
            //}
            SumDuration += M3uTable[M3uTableGetIndex].duration;
            DateTime nowdt = DateTime.Now;
            System.Diagnostics.Debug.WriteLine("M3uTableGetIndex=" + M3uTableGetIndex.ToString() + " 前回GETからの経過時間=" + (nowdt - LastGetDt).Seconds.ToString());
            LastGetDt = DateTime.Now;

            if (sndStream2 != null) {
                try {
                    sndStream2.Close();
                } catch (Exception) {
                }
            }
            String frgfile = ROOT_DIR + String.Format("{0:D3}", FragmentNo++);
            if (CurrentStation.transport_format == TRANSPORT_FORMAT.MP3) {
                frgfile += ".mp3";
            } else if (CurrentStation.transport_format == TRANSPORT_FORMAT.AAC) {
                frgfile += ".aac";
            } else if (CurrentStation.transport_format == TRANSPORT_FORMAT.TS_AAC) {
                frgfile += ".ts";
            }
            sndStream2 = new FileStream(frgfile, FileMode.Create);

            PcrStart = false;
        }


        private void ConnectAndFirstGet(String url)
        {
            CurrentUrl = new UrlInfo();

            System.Diagnostics.Debug.WriteLine("Connect");

            String[] urlpart = DevideUrl(url);
            CurrentUrl.host = urlpart[0];
            CurrentUrl.port = urlpart[1];
            CurrentUrl.file = urlpart[2];
            tcp = new TcpClient(CurrentUrl.host, Int32.Parse(CurrentUrl.port));
            stream = tcp.GetStream();
            SendGetRequest(CurrentUrl.host, CurrentUrl.file);
        }

        private void ReconnectAndFirstGet(String url)
        {
            String[] urlpart = DevideUrl(url);
            if (CurrentUrl != null && (CurrentUrl.host != urlpart[0] || CurrentUrl.port != urlpart[1])) {
                // HOSTまたはPORTが別
                Disconnect();

                ConnectAndFirstGet(url);
            } else {
                CurrentUrl.file = urlpart[2];
                SendGetRequest(CurrentUrl.host, CurrentUrl.file);
            }
        }

        private void Disconnect()
        {
            try {
                stream.Close();
                tcp.Close();
                tcp.Dispose();

                CurrentUrl = null;
            } catch (Exception ex) {

            }
        }

        /// <summary>
        /// URL(http://host:port/file)を, host,port,/fileの三つの部分に分割して返す。port指定されていない場合は"80"とする。
        /// 引数がURLとみなせない場合は、host,portは空白、引数全体を fileとする。
        /// </summary>
        /// <param name="url"></param>
        /// <returns></returns>
        private String[] DevideUrl(String url)
        {
            int idx0, idx1, idx2;
            String[] ret = new string[3];

            // https://host:port/file

            idx0 = url.IndexOf("://");
            if (idx0 > 0) {
                //if (String.Equals("http://", url.Substring(0, 7), StringComparison.CurrentCultureIgnoreCase) == true) {
                idx1 = url.IndexOf(':', idx0 + 3);
                if (idx1 > 0) {
                    ret[0] = url.Substring(idx0 + 3, idx1 - (idx0 + 3));
                    idx2 = url.IndexOf('/', idx1 + 1);
                    if (idx2 > 0) {
                        ret[1] = url.Substring(idx1 + 1, idx2 - idx1 - 1);
                        ret[2] = url.Substring(idx2);
                    } else {
                        ret[1] = url.Substring(idx1 + 1);
                        ret[2] = "/";
                    }
                } else {
                    idx1 = url.IndexOf('/', idx0 + 3);
                    if (idx1 > 0) {
                        ret[0] = url.Substring((idx0 + 3), idx1 - (idx0 + 3));
                        ret[1] = "80";
                        ret[2] = url.Substring(idx1);
                    } else {
                        ret[0] = url.Substring(idx0 + 3);
                        ret[1] = "80";
                        ret[2] = "/";
                    }
                }
            } else {
                ret[0] = "";
                ret[1] = "";
                ret[2] = url;
            }

            return ret;
        }


        /// <summary>
        /// 停止
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnStop_Click(object sender, EventArgs e)
        {
            Disconnect();
            //try {
            //    outStream.Close();
            //    sndStream.Close();
            //    sndtsStream.Close();
            //} catch (Exception ex) {
            //}
            DateTime dt = DateTime.Now;
            lblStopTime.Text = dt.ToString("HH:mm:ss");
            TimeSpan tspan = dt - StartDt;
            lblTimeElapse.Text = ((int)(tspan.TotalSeconds)).ToString();   // tspan.Minutes.ToString() + ":" + tspan.Seconds.ToString();
            lblSumDuration.Text = String.Format("{0}", SumDuration/1000);
            ReceiveStart = false;
        }


        private byte[] ResponseStatus(byte[] response)
        {
            byte[] ret = new byte[3];       // 
            int i = 0;
            int length = response.Length;

            do {
                if (response[i] == 0x20) break;
            } while (true);

            ret[0] = response[i + 1];
            ret[1] = response[i + 2];
            ret[2] = response[i + 3];

            return ret;
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            int i;

            for (i = 0; i < M3U_TABLE_SIZE; i++) {
                M3uTable[i] = new M3Utable();
            }
        }

        /// <summary>
        /// 188バイトに切り出されたTSパケットを解析
        /// </summary>
        /// <param name="fileposition"></param>
        private void ParseTsWork(long fileposition)
        {
            UInt16 pid;
            int section_length = 0;
            int program_number = 0;
            int adaptation_field_control = 0;
            int adaptation_field_length = 0;
            int payload_unit_start_indicator = 0;
            int transport_error_indecator = 0;
            int continuity_counter = 0;
            int discontinuity_indicator = 0;
            int pl_start;       // payload開始位置
            int payload_byte;
            int i;
            ulong pcr_base;
            ulong pcr_ext;
            ulong pcr_sec;
            int pcr_h, pcr_m, pcr_s;

            pl_start = 4;
            // TSヘッダ解析
            pid = (UInt16)(((TsWork[1] & 0x1F) << 8) | TsWork[2]);
            adaptation_field_control = ((TsWork[3] & 0x30) >> 4);
            transport_error_indecator = ((TsWork[1] & 0x80) >> 7);
            payload_unit_start_indicator = ((TsWork[1] & 0x40) >> 6);
            continuity_counter = (TsWork[3] & 0x0F);
            //System.Diagnostics.Debug.WriteLine(String.Format("{0:X5}", fileposition) +
            //                                " PID=" + pid.ToString() + " : " + String.Format("0x{0:x}", pid) +
            //                                " adaptation_field_control=" + String.Format("{0:X}", adaptation_field_control)
            //                                );
            //if (pid == TsElementaryPidSound) {
            //System.Diagnostics.Debug.WriteLine("HEADER sync={0:X2} pid={1:X4} cont_cnt={2:X2} adp_f_c={3} tei={4} pusi={5}", 
            //    TsWork[0], pid, continuity_counter, adaptation_field_control, transport_error_indecator, payload_unit_start_indicator);
            //}

            // Adaptation field
            if ((adaptation_field_control & 0x02) != 0) {       // Adaptation field あり
                adaptation_field_length = TsWork[pl_start];
                discontinuity_indicator = ((TsWork[pl_start + 1] & 0x80) >> 7);
                if ((TsWork[pl_start + 1] & 0x10) != 0) {       // PCR flag
                    ulong pcr_last = Pcr;
                    ulong diff_pcr;
                    int diff_pcr_msec;
                    int tc_pcr_last = TickCountAtPcr;

                    //if (PcrStart == false 
                    //    || (PcrStart == true && TsWork[pl_start + 2] == PcrByte[0] && TsWork[pl_start + 3] == PcrByte[1])
                    //    || (PcrStart == true && TsWork[pl_start + 2] == PcrByte[0] && TsWork[pl_start + 3] == (PcrByte[1] + 1))
                    //    || (PcrStart == true && TsWork[pl_start + 2] == 0 && TsWork[pl_start + 3] == 0)
                    //    ) {     // 有効な？PCR値のとき

                        pcr_base = (((ulong)TsWork[pl_start + 2] << (3 * 8)) & 0x00000000FF000000);
                        pcr_base |= (((ulong)TsWork[pl_start + 3] << (2 * 8)) & 0x0000000000FF0000);
                        pcr_base |= (((ulong)TsWork[pl_start + 4] << (1 * 8)) & 0x000000000000FF00);
                        pcr_base |= (((ulong)TsWork[pl_start + 5] << (0 * 8)) & 0x00000000000000FF);
                        pcr_base <<= 1;
                        pcr_base |= (((ulong)TsWork[pl_start + 6] >> (7)) & 0x0000000000000001);
                        pcr_ext = (((ulong)TsWork[pl_start + 6] << (8)) & 0x0000000000000100);
                        pcr_ext |= (((ulong)TsWork[pl_start + 7] << (0)) & 0x00000000000000FF);
                        Pcr = pcr_base * 300 + pcr_ext;
                        pcr_sec = Pcr / 27000000;               // PCRを秒換算
                        pcr_h = (int)(pcr_sec / 3600);
                        pcr_m = (int)((pcr_sec % 3600) / 60);
                        pcr_s = (int)(pcr_sec % 60);
                        TickCountAtPcr = Environment.TickCount;

                        if (PcrStart == true) {
                            if (Pcr > pcr_last) {
                                diff_pcr = Pcr - pcr_last;
                            } else {
                                diff_pcr = 0;   // TODO:
                            }
                            diff_pcr_msec = (int)((diff_pcr * 1000) / 27000000);   // diff_pcrをmsec換算
                            SumPcrMsec += diff_pcr_msec;
                            SumTcMsec += TickCountAtPcr - tc_pcr_last;
                            System.Diagnostics.Debug.WriteLine("PID=" + pid.ToString()
                                + " trns_err_ind=" + transport_error_indecator.ToString()
                                + " pyld_unt_st_ind=" + payload_unit_start_indicator.ToString()
                                + " discnt_ind=" + discontinuity_indicator.ToString()
                                + " PCR=" + String.Format("{0}", Pcr)
                                + " " + String.Format("{0:d2}:{1:d2}:{2:d2}", pcr_h, pcr_m, pcr_s)
                                //+ " diff(clock)=" + String.Format("{0}", diff_pcr)
                                //+ " diff(msec)=" + String.Format("{0}", diff_pcr_msec)
                                //+ " SUM(msec)=" + String.Format("{0}", SumPcrMsec)
                                //+ " tc=" + String.Format("{0}", TickCountAtPcr - tc_pcr_last)
                                //+ " SUM(msec)=" + String.Format("{0}", SumTcMsec)
                                );
                            //System.Diagnostics.Debug.WriteLine(" {0:X2} {1:X2} {2:X2} {3:X2} {4:X2} {5:X2} ",
                            //    TsWork[pl_start + 2], TsWork[pl_start + 3], TsWork[pl_start + 4], TsWork[pl_start + 5], TsWork[pl_start + 6], TsWork[pl_start + 7]);
                        } else {
                            PcrStart = true;
                            SumPcrMsec = 0;
                            SumTcMsec = 0;
                        }

                        Array.Copy(TsWork, pl_start + 2, PcrByte, 0, 6);
                    //} else {
                        //System.Diagnostics.Debug.WriteLine(" {0:X2} {1:X2} {2:X2} {3:X2} {4:X2} {5:X2} ",
                        //    TsWork[pl_start + 2], TsWork[pl_start + 3], TsWork[pl_start + 4], TsWork[pl_start + 5], TsWork[pl_start + 6], TsWork[pl_start + 7]);
                    //}
                }
                //if ((TsWork[pl_start + 1] & 0x08) != 0) {
                //    System.Diagnostics.Debug.WriteLine(" OPCR exists");
                //}
                //if ((TsWork[pl_start + 1] & 0x04) != 0) {
                //    System.Diagnostics.Debug.WriteLine(" スプライシング exists");
                //}
                //if ((TsWork[pl_start + 1] & 0x02) != 0) {
                //    System.Diagnostics.Debug.WriteLine(" transport private data exists");
                //}
                //if ((TsWork[pl_start + 1] & 0x01) != 0) {
                //    System.Diagnostics.Debug.WriteLine(" adaptation field extention exists");
                //}
                pl_start += (adaptation_field_length + 1);
                //System.Diagnostics.Debug.WriteLine(" adaptation_field_length=" + String.Format("{0:D}", adaptation_field_length));
            }
            if ((adaptation_field_control & 0x01) != 0) {       // Payloadあり
                payload_byte = TS_SIZE - pl_start;
            } else {        // Payloadなし
                payload_byte = 0;
            }

            //if (payload_byte > 0 ) {
            //    System.Diagnostics.Debug.WriteLine("Pointer field=" + TsWork[pl_start]);
            //}

            if (pid == 0) {     // PAT
                                // payloadの先頭1バイトはpointer field 。そのあとからPAT(table_id, section_syntax_indicator,...,section_length, ...)
                section_length = ((TsWork[pl_start + 2] & 0x0F) << 8) | TsWork[pl_start + 3];
                // 
                i = 9;
                program_number = (TsWork[pl_start + i] << 8 | TsWork[pl_start + i + 1]);
                TsPmtPid = ((TsWork[pl_start + i + 2] & 0x1F) << 8) | TsWork[pl_start + i + 3];
                System.Diagnostics.Debug.WriteLine("PAT section_length=" + section_length
                    + String.Format(" PAT program_number={0}(0x{1:x}), pmt_pid={2}(0x{3:x})", program_number, program_number, TsPmtPid, TsPmtPid));
            } else if (pid == TsPmtPid && pid != PID_NULL) {        // PMT
                int program_info_length = 0;
                int stream_type = 0;
                // payloadの先頭1バイトはpointer field 。そのあとからPMT(table_id, ....)
                section_length = ((TsWork[pl_start + 2] & 0x0F) << 8) | TsWork[pl_start + 3];
                program_info_length = ((TsWork[pl_start + 11] & 0x0F) << 8) | TsWork[pl_start + 12];

                stream_type = TsWork[pl_start + 13 + program_info_length];

                TsElementaryPidSound = ((TsWork[pl_start + 13 + program_info_length + 1] & 0x1F) << 8) | TsWork[pl_start + 13 + program_info_length + 2];
                System.Diagnostics.Debug.WriteLine("PMT " + "pl_start=" + pl_start + " section_length = " + section_length + "  program_info_length=" + program_info_length
                    + " stream_type=" + stream_type
                    + " elementary_pid=" + TsElementaryPidSound);

            } else if (pid == TsElementaryPidSound) {
                //System.Diagnostics.Debug.WriteLine("PES ");
                //System.Diagnostics.Debug.WriteLine(String.Format("{0:X5}", fileposition + pl_start) + " Audio " + String.Format("{0}", payload_byte));
                sndtsStream.Write(TsWork, pl_start, TS_SIZE - pl_start);
            } else if (pid != PID_NULL){
                System.Diagnostics.Debug.WriteLine("OTHER PID=" + pid);
            }
        }

        /// <summary>
        /// TSファイルを解析。音源データだけ抽出する。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnAnalizeTs_Click(object sender, EventArgs e)
        {
            byte[] buf = new byte[256];
            UInt16 pid;
            int section_length = 0;
            int program_number = 0;
            int adaptation_field_control = 0;
            int adaptation_field_length = 0;
            int pmt_pid = 0;
            int elementary_pid = 0;
            int i;
            bool continue_flag = true;
            long fileposition = 0;
            int pl_start;       // payload開始位置
            int payload_byte;

            sndStream = new FileStream(DATFILE_TS, FileMode.Open);
            sndStream2 = new FileStream(SNDFILE2, FileMode.Create);
            try {
                // TS header
                while (continue_flag) {
                    // TSヘッダの先頭バイト(sync_byte)取得
                    do {
                        fileposition = sndStream.Position;
                        if (sndStream.Read(buf, 0, 1) < 1) {
                            continue_flag = false;
                            break;
                        }
                        if (buf[0] == 0x47) {
                            break;
                        }
                        System.Diagnostics.Debug.WriteLine(String.Format(" sync byte 不正 (POS=0x{0:X})(data=0x{1:X})", fileposition, buf[0]));
                    } while (true);
                    if (continue_flag == false) break;

                    // パケットの残り取得
                    if (sndStream.Read(buf, 1, 187) < 187) break;
                    pl_start = 4;
                    // TSヘッダ解析
                    pid = (UInt16)(((buf[1] & 0x1F) << 8) | buf[2]);
                    adaptation_field_control = ((buf[3] & 0x30) >> 4);
                    //System.Diagnostics.Debug.WriteLine(String.Format("{0:X5}", fileposition) +
                    //                                " PID=" + pid.ToString() + " : " + String.Format("0x{0:x}", pid) +
                    //                                " adaptation_field_control=" + String.Format("{0:X}", adaptation_field_control)
                    //                                );
                    // Adaptation field
                    if ((adaptation_field_control & 0x02) != 0) {       // Adaptation field あり
                        adaptation_field_length = buf[pl_start];
                        pl_start += (adaptation_field_length + 1);
                        System.Diagnostics.Debug.WriteLine(" adaptation_field_length=" + String.Format("{0:D}", adaptation_field_length));
                    }
                    if ((adaptation_field_control & 0x01) != 0) {       // Payloadあり
                        payload_byte = 188 - pl_start;
                    } else {        // Payloadなし
                        payload_byte = 0;
                    }

                    if (pid == 0) {     // PAT
                        // payloadの先頭1バイトはpointer field 。そのあとからPAT(table_id, section_syntax_indicator,...,section_length, ...)
                        section_length = ((buf[pl_start + 2] & 0x0F) << 8) | buf[pl_start + 3];
                        System.Diagnostics.Debug.WriteLine("PAT section_length=" + section_length);
                        // 
                        i = 9;
                        program_number = (buf[pl_start + i] << 8 | buf[pl_start + i + 1]);
                        pmt_pid = ((buf[pl_start + i + 2] & 0x1F) << 8) | buf[pl_start + i + 3];
                        System.Diagnostics.Debug.WriteLine(String.Format("PAT program_number={0}(0x{1:x}), pmt_pid={2}(0x{3:x})", program_number, program_number, pmt_pid, pmt_pid));
                    } else if (pid == pmt_pid) {        // PMT
                        int program_info_length = 0;
                        int stream_type = 0;
                        // payloadの先頭1バイトはpointer field 。そのあとからPMT(table_id, ....)
                        section_length = ((buf[pl_start + 2] & 0x0F) << 8) | buf[pl_start + 3];
                        System.Diagnostics.Debug.WriteLine("PMT section_length=" + section_length);
                        program_info_length = ((buf[pl_start + 11] & 0x0F) << 8) | buf[pl_start + 12];
                        System.Diagnostics.Debug.WriteLine("PMT program_info_length=" + program_info_length);

                        stream_type = buf[pl_start + 13 + program_info_length];
                        System.Diagnostics.Debug.WriteLine("PMT stream_type=" + stream_type);

                        elementary_pid = ((buf[pl_start + 13 + program_info_length + 1] & 0x1F) << 8) | buf[pl_start + 13 + program_info_length + 2];
                        System.Diagnostics.Debug.WriteLine("PMT elementary_pid=" + elementary_pid);
                    } else if (pid == elementary_pid) {
                        System.Diagnostics.Debug.WriteLine(String.Format("{0:X5}", fileposition + pl_start) + " Audio " + String.Format("{0}", payload_byte));
                        sndStream2.Write(buf, pl_start, 188 - pl_start);
                    }
                }

            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
            } finally {
                sndStream.Close();
                sndStream2.Close();
            }

        }

        /// <summary>
        /// host, port, requestを個別に指定してテスト
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void txtConnect2_Click(object sender, EventArgs e)
        {
            TcpClient tcp2;
            NetworkStream stream2;

            tcp2 = new TcpClient(txtHost.Text, Int32.Parse(txtPort.Text));
            stream2 = tcp2.GetStream();

            Byte[] sendBytes = Encoding.ASCII.GetBytes(txtGet.Text);    // Encoding.UTF8.GetBytes(msg);
            stream2.Write(sendBytes, 0, sendBytes.Length);

            while (stream2.DataAvailable == false) {

            }
            while (stream2.DataAvailable == true) {
                int read_byte = stream2.Read(RecvBuf, 0, RECV_BUF_SIZE);
                String response = Encoding.UTF8.GetString(RecvBuf, 0, read_byte);
                System.Diagnostics.Debug.WriteLine(response);
            }
        }
    }
}
